#
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#

package CModeAgent;
use NaServer;
use NaElement;
use ArrayRep;
@ISA = ("ArrayRep");
use strict;
use ag_i18n_inc;
#use  Exporter;
#@EXPORT = (
#'IsClusteredEnvironment'
#);
sub new {
	my ($self, $resname, $filername, $lang, $cluster, $vserver, $volume) = @_;
	my $class = ref($self) || $self;
	my $filer = {};

	VCSAG_LOGDBG_MSG(4, "sub new");
	$filer->{START_TIME} = $^T;
	$filer->{resname} = $resname;
	$filer->{LANG} = $lang;
	$filer->{FILER_NAME} = $filername;
	$filer->{SNAPMIRROR_ON} = "false";
	$filer->{CLUS_NAME} = $cluster;
	$filer->{VSERVER_NAME} = $vserver;
	$filer->{VOLUME} = $volume;
	VCSAG_LOGDBG_MSG(4, "volume=  $filer->{VOLUME}, Cmode cluster = $filer->{CLUS_NAME}, Vserver = $filer->{VSERVER_NAME}");
	
	return bless $filer, $class;
}

sub new_filer {
	my ($self, $resname, $filername, $lang, $type) = @_;
	my $class = ref($self) || $self;
	my $filer = {};

	VCSAG_LOGDBG_MSG(4, "sub new_filer()");
	$filer->{START_TIME} = $^T;
	$filer->{resname} = $resname;
	$filer->{LANG} = $lang;
	$filer->{FILER_NAME} = $filername;
	$filer->{resource_type} = $type;
	VCSAG_LOGDBG_MSG(3, "new_filer() called for resource [$filer->{resname}]");
	
	return bless $filer, $class;
}

sub set_xml_debug_style {
	my $self = shift;
	my $debug_style = "NA_PRINT_DONT_PARSE";
	my $handle = $self->{FILER_HANDLE};
	$handle->set_debug_style($debug_style);
	return;
}

#Sub check for filer configuration
#Return 0 : for 7 Mode configuration
#Return 1: for CMode configuration
#Return 2: for failure scenario
sub IsClusteredEnvironment {
	my ($filername, $user, $enc_pword) = @_;
	my $is_clus = "";
	VCSAG_LOGDBG_MSG(1, "sub IsClusteredEnvironment");
        my $handle = NaServer->new($filername, 1, 1);

        $enc_pword =~s/^\s+//;
        $enc_pword =~s/\s+$//;
        if ($enc_pword) {
                my $Pword = `$HADECRYPT $enc_pword`;
                if ($? != 0) {
                        VCSAG_LOG_MSG("W", "Encountered errors while decrypting password!", 1059);
                        $Pword = "";
                }
                $handle->set_admin_user($user, $Pword);
        } else {
                $handle->set_style("HOSTS");
        }

        my $output = $handle->invoke("system-get-version");
        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                VCSAG_LOG_MSG("E", "ONTAPI 'system-get-version' failed on filer $filername. Error $output->{'results_errno'}: $r", 1060, "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 2;
        }

	$is_clus = $output->child_get_string( "is-clustered" );
        if($is_clus eq "true"){
                VCSAG_LOG_MSG("I", "Configuration is Clustered Data ONTAP. Filer Name: $filername ", 1061, "$filername");
                return 1;
        }

        VCSAG_LOG_MSG("I", "Configuration is 7-Mode. Filer Name: $filername ", 1062, "$filername");
        return 0;
	
}

sub OpenConnection {
        my ($self, $user, $enc_pword) = @_;
	VCSAG_LOGDBG_MSG(1, "sub OpenConnection");
        my $filername = $self->{FILER_NAME};

        my $handle = NaServer->new($filername, 1, 1);

        $enc_pword =~s/^\s+//;
        $enc_pword =~s/\s+$//;
        
        # if password is specified, use it to login as root on the filer
        # if not, use the hosts.equiv file on filer to determine access rights.
        #                        #
        #
        if ($enc_pword) {
                my $Pword = `$HADECRYPT $enc_pword`;
		my $ret_decr = $?;
		if ($ret_decr != 0) {
			VCSAG_LOG_MSG("W", "Encountered errors while decrypting password!.Error[$ret_decr]", 1063);
			$Pword = "";
		}
                $handle->set_admin_user($user, $Pword);
        } else {
                $handle->set_style("HOSTS");
        }

        my $output = $handle->invoke("system-get-version");
        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'system-get-version' failed on filer $filername. Error $output->{'results_errno'}: $r", 1064, "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $version = $output->child_get_string( "version" );
        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOGDBG_MSG(1, "ONTAP version on filer $filername is: $version", "$filername", "$version");
        $ENV{"LANG"}="C";

        $self->{FILER_HANDLE} = $handle;

        return 0;
}

#
# This routine will wait in a tight loop for the snapmirror status
# of the specified volume to become 'idle'. This is required because
# certain snapmirror commands can only be issued when the status is
# idle or else the command will fail.
#
sub WaitForIdle {
        my ($self ) = @_;
	VCSAG_LOGDBG_MSG(1, "sub WaitForIdle");

        my @sminfolist;

        my $filername = $self->{FILER_NAME};
      # wait for snapmirror status to be "idle"

        while (1) {
                @sminfolist = $self->SnapMirrorGetInfo();
                return 1 if (!@sminfolist);

                #
                #we expect to see only one entry in the snapmirror infolist.
                #this agent does not support multiple snapmirror destinations
                #for the same source volume
                #                                                                #
 
                if (scalar(@sminfolist) != 1) {
                        $ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for filer $filername returned multiple status entries. Administrative intervention required", 1065,  "$filername");
                        $ENV{"LANG"}="C";
                        return 1;
                }
                my $sm_info = $sminfolist[0];
                last if ($sm_info->{STATUS} eq "idle");
                undef(@sminfolist);
                # sleep 5 seconds before checking again
               sleep 5;
			
	}
	return 0;
}

sub get_state {
        my ($self, $location) = @_;
	VCSAG_LOGDBG_MSG(1, "sub get_state");

        my $filername = $self->{FILER_NAME};
        my( @sminfolist ) = $self->SnapMirrorGetInfo();
        if (scalar(@sminfolist) > 1) {
                $ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOGDBG_MSG(1, "multiple instance returned.");
                VCSAG_LOG_MSG("E", "Multiple snapmirror for a volume is not supported by this agent. 'snapmirror get' for volume $location on filer $filername returned multiple status entries. Administrative intervention required", 1066, "$location", "$filername");
                $ENV{"LANG"}="C";
                return undef;
        }

        return $sminfolist[0];
}

sub SnapMirrorGetRemoteInfo {
        my ($self, $location) = @_;
        VCSAG_LOGDBG_MSG(1, "sub SnapMirrorGetRemoteInfo");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};

        VCSAG_LOGDBG_MSG(1, "destination-location[$location]");

        my $output = $handle->invoke("snapmirror-get", "destination-location" ,$location);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-get' failed for volume $location on filer $filername. Error $output->{'results_errno'}: $r", 1067, "$location", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
        	VCSAG_LOGDBG_MSG(1, "destination-location[$location] on filer $filername. Error $output->{'results_errno'}: $r ");
                return undef;
        }

         my $statuslist = $output->child_get("attributes");
      if (!defined($statuslist) || ($statuslist eq "")) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Volume $location on filer $filername is not a snapmirror replicated volume", 1068, "$location", "$filername");
                $ENV{"LANG"}="C";
                return undef;
        }


        my @infolist = $statuslist->children_get();
        my @sminfolist = ();

        foreach my $sm_info (@infolist) {
                my $info = {};
                $info->{STATE} = $sm_info->child_get_string("mirror-state");
                $info->{STATUS} = $sm_info->child_get_string("relationship-status");

                $info->{DST_LOC} = $sm_info->child_get_string("destination-location");
                $info->{DST_VOL} = $sm_info->child_get_string("destination-volume");
                $info->{DEST_V_SER} = $sm_info->child_get_string("destination-vserver");
                $info->{SRC_LOC} = $sm_info->child_get_string("source-location");
                $info->{SRC_VOL} = $sm_info->child_get_string("source-volume");
                $info->{SRC_V_SER} = $sm_info->child_get_string("source-vserver");
                $info->{VSERVER} = $sm_info->child_get_string("vserver");
                $info->{SCHEDULE} = $sm_info->child_get_string("schedule");
                VCSAG_LOGDBG_MSG(2, "Remote mirror-state = $info->{STATE}  relationship-status = $info->{STATUS}");
                VCSAG_LOGDBG_MSG(2,"Remote source-location = $info->{SRC_LOC} source-volume = $info->{SRC_VOL}");
                VCSAG_LOGDBG_MSG(2, "Remote vserver = $info->{VSERVER} source-vserver = $info->{SRC_V_SER}");
                VCSAG_LOGDBG_MSG(2, "Remote schedule = $info->{SCHEDULE}");
                VCSAG_LOGDBG_MSG(1, "Remote destination-volume = $info->{DST_VOL}");
                VCSAG_LOGDBG_MSG(4, "Remote destination-location = $info->{DST_LOC} destination-vserver= $info->{DEST_V_SER}");

                push(@sminfolist, $info);
        }

       return (@sminfolist);

}

sub SnapMirrorGetInfo {
        my ($self) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorGetInfo");

	my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
	my $vserver = $self->{VSERVER_NAME};
	my $clus_name = $self->{CLUS_NAME};
	my $location =  $self->{VOLUME};
	my $output;

	if ( !length($clus_name) ) {
		VCSAG_LOGDBG_MSG(1, "destination-volume[$location], VSERVER_NAME = $vserver");
     		$output = $handle->invoke("snapmirror-get", "destination-volume" ,$location, "destination-vserver", $vserver);
	}else {    
		VCSAG_LOGDBG_MSG(1, "destination-volume[$location], CLUS_NAME = $clus_name, VSERVER_NAME = $vserver");
     		$output = $handle->invoke("snapmirror-get", "destination-volume" ,$location, "destination-cluster", $clus_name, "destination-vserver", $vserver);
	}

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("W", "ONTAPI 'snapmirror-get' failed for volume $location on filer $filername. Error $output->{'results_errno'}: $r", 1069, "$location", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return undef;
        }

	# get snapmirror status	<mirror-state>snapmirrored</mirror-state> 
	 my $statuslist = $output->child_get("attributes");
      	if (!defined($statuslist) || ($statuslist eq "")) {
              # no snapmirror status to report
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Volume $location on filer $filername is not a snapmirror replicated volume", 1070, "$location", "$filername");
                $ENV{"LANG"}="C";
                return undef;
	}
        
	# iterate through snapmirror status info list.
        # NOTE: THIS AGENT CURRENTLY SUPPORTS ONLY ONE DESTINATION PER
        # SOURCE VOLUME
        #

        my @infolist = $statuslist->children_get();
        my @sminfolist = ();

	VCSAG_LOGDBG_MSG(1, "****** Information from snapmirror-get*********");
        foreach my $sm_info (@infolist) {
                my $info = {};
                $info->{STATE} = $sm_info->child_get_string("mirror-state");
                $info->{STATUS} = $sm_info->child_get_string("relationship-status");
		
		$info->{DST_LOC} = $sm_info->child_get_string("destination-location");	
		$info->{DST_VOL} = $sm_info->child_get_string("destination-volume");	
		$info->{DEST_V_SER} = $sm_info->child_get_string("destination-vserver");	
		$info->{SRC_LOC} = $sm_info->child_get_string("source-location");	
		$info->{SRC_VOL} = $sm_info->child_get_string("source-volume");	
		$info->{SRC_V_SER} = $sm_info->child_get_string("source-vserver");	
		$info->{VSERVER} = $sm_info->child_get_string("vserver");	
       		$info->{SCHEDULE} = $sm_info->child_get_string("schedule");	
		$info->{POLICY} =  $sm_info->child_get_string("policy");
		$info->{HEALTHY} = $sm_info->child_get_string("is-healthy");
		VCSAG_LOGDBG_MSG(1, "mirror-state = $info->{STATE}  relationship-status = $info->{STATUS}");
		VCSAG_LOGDBG_MSG(2, "destination-location = $info->{DST_LOC}, destination-vserver= $info->{DEST_V_SER}");		 VCSAG_LOGDBG_MSG(1,"source-location = $info->{SRC_LOC}, source-volume = $info->{SRC_VOL}");
		VCSAG_LOGDBG_MSG(2, "vserver = $info->{VSERVER}, source-vserver = $info->{SRC_V_SER}");
		VCSAG_LOGDBG_MSG(1, "schedule = $info->{SCHEDULE}, policy = $info->{POLICY}");
		VCSAG_LOGDBG_MSG(1, "destination-volume = $info->{DST_VOL}");
		VCSAG_LOGDBG_MSG(1, "is-healthy = $info->{HEALTHY}");

                push(@sminfolist, $info);
	}

	VCSAG_LOGDBG_MSG(1, "****** End information from snapmirror-get*********");
       return  (@sminfolist);

}

sub failover_required {
        my ($self, $sm_info) = @_;
	VCSAG_LOGDBG_MSG(1, "sub failover_required");
	VCSAG_LOGDBG_MSG(1, "STATE = $sm_info->{STATE}");
        return 2 if ($sm_info->{STATE} eq "broken-off");
        return 1 if ($sm_info->{STATE} eq "snapmirrored");

        #in 8.2 and 8.3 no snapmirror information found on source site
        #check if volume is already in read write(rw) type  if yes assume its source site and mark resource online
        #
	my $retcode = $self->GetVolumeType();
	if(!$retcode) {
        	VCSAG_LOGDBG_MSG(1, "volume type is RW ");
                return 0;
	}

        return -1;
}

sub SnapMirrorUpdate {
        my ($self, $dstloc, $wait) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorUpdate");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};

        # wait for snapmirror status to be "idle"
        my $ret = $self->WaitForIdle();
        if ($ret) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Error while waiting for snapmirror status of volume $dstloc to become 'idle' on filer $filername", 1072, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

	VCSAG_LOGDBG_MSG(3, "API: snapmirror-update destination-location $dstloc");
        my $output = $handle->invoke("snapmirror-update",
                "destination-location", $dstloc);
	VCSAG_LOGDBG_MSG(3, "API Ret: snapmirror-update destination-location $dstloc");

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-update' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1073, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror update initiated for volume $dstloc", 1028, "$dstloc");
        $ENV{"LANG"}="C";

        return 0 unless ($wait);

	# wait for update to complete or (TBD: timeout)
        
        $ret = $self->WaitForIdle();
        if ($ret) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Error while waiting for snapmirror update of volume $dstloc to complete on filer $filername", 1074, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }
        return 0;
}


sub SnapMirrorQuiesce {
 	my ($self, $dstloc ) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorQuiesce");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
        my ( $output, $r );
        my $quiesce_retries = 3;
        my $quiesce_interval = 5;
        my $quiesce_failure = 0;

        #
        #wait for status to become idle or (TBD: timeout)
        #                
        
        my $ret = $self->WaitForIdle();
        if ($ret) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Error while waiting for snapmirror status of volume $dstloc to become 'idle' on filer $filername", 1075, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("N", "Attempting to quiesce snapmirror location $dstloc", 1076, "$dstloc");
        $ENV{"LANG"}="C";
        
        #Quiesce fails in some cases where there are
        #multiple volumes being simultaneously replicated. Retry
        #                        
        
        for ( my $ii = $quiesce_retries; $ii > 0; $ii-- ){
                $output = $handle->invoke("snapmirror-quiesce",
                        "destination-location", $dstloc);

                if ($output->results_errno != 0) {
                        $r = $output->results_reason();
                        $ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("W", "ONTAPI 'snapmirror-quiesce' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r. Retrying ...", 1079, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                        $ENV{"LANG"}="C";
                        $quiesce_failure = 1;
                }else {
                        $quiesce_failure = 0;
                        last;
                }
                sleep $quiesce_interval ;
        }

        if ( $quiesce_failure == 1 ){
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-quiesce' failed $quiesce_retries times for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r. Giving up", 1080, "$quiesce_retries", "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        #
        #check if volume was successfully quiesced
        #

        my @sminfolist = $self->SnapMirrorGetInfo();
        return 1 if (!@sminfolist);
                
        #
        #we expect to see only one entry in the snapmirror infolist.
        # this agent does not support multiple snapmirror destinations
        #for the same source volume
        #                                
       
        if (scalar(@sminfolist) != 1) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $filername returned multiple status entries. Administrative intervention required", 1081, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $sm_info = $sminfolist[0];
        if ($sm_info->{STATUS} ne "quiesced") {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "snapmirror-quiesce failed to quiesce the destination location $dstloc at filer $filername", 1082, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirrored location $dstloc has been successfully quiesced", 1083, "$dstloc");
        $ENV{"LANG"}="C";

        return 0;
}


# Breaks a snapmirrored relationship on this filer for specified volume.
# Also queries status afterward to see if the volume was successfully
# broken-off
#
#
sub SnapMirrorBreak {
 	my ($self, $dstloc) = @_;

	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorBreak");
        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("N", "Attempting to break snapmirror location $dstloc", 1023, "$dstloc");
        $ENV{"LANG"}="C";

        my $output = $handle->invoke("snapmirror-break",
                "destination-location", $dstloc);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-break' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1082, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        #
        # check if volume was successfully broken-off
        #

        my @sminfolist = $self->SnapMirrorGetInfo();
        return 1 if (!@sminfolist);

        #
        #we expect to see only one entry in the snapmirror infolist.
        #this agent does not support multiple snapmirror destinations
        #for the same source volume
        #
        
        if (scalar(@sminfolist) != 1) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $filername returned multiple status entries. Administrative intervention required", 1084, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $sm_info = $sminfolist[0];
        if ($sm_info->{STATE} ne "broken-off") {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "snapmirror-break failed to break snapmirror relationship for destination location $dstloc at filer $filername", 1085, "$dstloc", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirrored location $dstloc has been successfully broken-off", 1086, "$dstloc");
        $ENV{"LANG"}="C";

        return 0;
}



sub migrate {
        my ($lfiler, $rfiler, $sm_info, $schedule, $ops_throttle, $visibility) = @_;
	VCSAG_LOGDBG_MSG(1, "sub migrate");

        #
        # volume on lfiler filer is the current destination (to be made source)
        # volume on remote filer is the current source (to be made destination)
        #

        my $dstloc = $sm_info->{DST_LOC};
        my $srcloc = $sm_info->{SRC_LOC};
        my $sched_exists = $sm_info->{SCHEDULE};
	my $policy_exists = $sm_info->{POLICY};
        my $srcloc_to_config = $dstloc;
        my $dstloc_to_config = $srcloc;

        my ($rflrname_conf, $rvolname_conf) = split(/:/, $srcloc);
        my ($lflrname, $lvolname) = split(/:/, $dstloc);
        my $rflrname = $rfiler->{FILER_NAME};

       
        my $ret = $lfiler->SnapMirrorUpdate($dstloc, 1);
        if ($ret) {
                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to update snapmirror location $dstloc on filer $lflrname", 1087, "$dstloc", "$lflrname");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ret = $lfiler->ISReplicationHealthy();
        if ($ret) {

                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to check is-healthy attribute for replication. Possible reason replicatin is not in healthy state or multiple entry found. Admin intervention is required. Filer $lflrname", 1120, "$lflrname");
                $ENV{"LANG"}="C";
                return 1;
        }


        $ret = $lfiler->SnapMirrorQuiesce($dstloc);
        if ($ret) {
                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to queisce destination location $dstloc on filer $lflrname", 1088, "$dstloc", "$lflrname");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ret = $lfiler->SnapMirrorBreak($dstloc);
        if ($ret) {
                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to break destination location $dstloc on filer $lflrname", 1089, "$dstloc", "$lflrname");
                $ENV{"LANG"}="C";
                return 1;
        }
     #check replication is-healthy again . Sometime is-healthy goes into false after replication break operation.

        $ret = $lfiler->ISReplicationHealthy();
        if ($ret) {

                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "After replication break failed to check is-healthy attribute. Possible reason replicatin is not in healthy state or multiple entry found. Admin intervention is required. Filer $lflrname", 1121, "$lflrname");
                $ENV{"LANG"}="C";
                return 1;
        }

        #
        # Resync new destination (i.e remote filer) with new
        # source (i.e local filer)
        #                        #
        $ret = $rfiler->SnapMirrorResync($dstloc_to_config, $srcloc_to_config);
        if ($ret) {
                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to resync new destination location $dstloc_to_config on filer $rflrname new source location $srcloc_to_config", 1090, "$srcloc_to_config", "$rflrname", "$srcloc_to_config");
                $ENV{"LANG"}="C";
                return 1;
        }
	
        my $done = 0;
        while (!$done) {
		sleep 1;
                my @sminfolist = $rfiler->SnapMirrorGetRemoteInfo($dstloc_to_config);
                return 1 if (!@sminfolist);

                # find the right snapmirror-info entry in the infolist and
                # wait till Status goes to Idle (there should ideally be only
                # 2 entries in the list)
                # TBD: we need to consider situations where status goes to
                # something other that 'idle' after a resync is initiated.
                
                foreach my $sm_info (@sminfolist) {
                        next if ($sm_info->{DST_LOC} ne $dstloc_to_config);

                        if ($sm_info->{STATUS} eq "idle") {
                                $done = 1;
                        } else {
                                sleep 5;
                                $ENV{"LANG"}="$lfiler->{LANG}";
                                VCSAG_LOG_MSG("N", "Waiting for snapmirror resync to destination location $srcloc on filer $rflrname to complete", 1091, "$srcloc", "$rflrname");
                                $ENV{"LANG"}="C";
                        }
                        last;
                }
        }


	#
        # Find and delete base-snapshot from local filer (i.e original dest)
        #
        VCSAG_LOGDBG_MSG(1, "Call SnapMirrorGetInfo() to delete base-snapsot from local filer");
        my @sminfolist = $lfiler->SnapMirrorGetInfo();
        return 1 if (!@sminfolist);
 
	$done = 0;
        foreach my $sm_info (@sminfolist) {
                if ($sm_info->{DST_LOC} eq $dstloc) {
                        $ret = $lfiler->SnapshotDelete($dstloc);
                        if ($ret) {
                                $ENV{"LANG"}="$lfiler->{LANG}";
                                VCSAG_LOG_MSG("E", "Failed to destroy  snapshot for location $dstloc on filer $lflrname", 1092, "$dstloc", "$lflrname");
                                $ENV{"LANG"}="C";
                                return 1;
                        }
                        $done = 1;
                        last;
                }
        }

        if( $sched_exists ne "") {
		$ret = $rfiler->SnapMirrorSetSchedule($dstloc_to_config, $sched_exists);
        	if ($ret) {
                	$ENV{"LANG"}="$lfiler->{LANG}";
                	VCSAG_LOG_MSG("E", "Failed to modify schedule: $sched_exists location: $dstloc on filer: $rflrname", 1093,"$sched_exists" ,"$dstloc", "$rflrname");
                	$ENV{"LANG"}="C";
                #TODO: TBD Do we required to return fail from here.
        	}
	}

        if(( $policy_exists ne "") && ( $policy_exists ne "DPDefault")) {
                $ret = $rfiler->SnapMirrorSetPolicy($dstloc_to_config, $policy_exists);
                if ($ret) {
                        $ENV{"LANG"}="$lfiler->{LANG}";
                        VCSAG_LOG_MSG("E", "Failed to modify policy: $policy_exists location: $dstloc on filer: $rflrname", 1094,"$policy_exists" ,"$dstloc", "$rflrname");
                        $ENV{"LANG"}="C";
			return 1;
                #TODO: TBD Do we required to return fail from here.
		}
        }
       
 
	$ret = $rfiler->SnapMirrorRelease($dstloc);
        if ($ret) {
                $ENV{"LANG"}="$lfiler->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to release previous snapmirror relationship between original source location $srcloc and original destination location $dstloc on filer $rflrname", 1095, "$srcloc", "$dstloc", "$rflrname");
                $ENV{"LANG"}="C";
                return 1;
        }

	return 0;
}

#
# This routine intiates a snapmirror resync. It is up to the caller
# to wait for resync to complete, if required.

sub SnapMirrorResync {
        my ($self, $dstloc, $srcloc) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorResync");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
        my ($output, $r);

        my $resync_retries = 50;
        my $resync_interval = 5;
        my $resync_failure = 0;

        VCSAG_LOG_MSG("N", "Initaiating snapmirror resync for destination location $dstloc source location: $srcloc", 1096, "$dstloc");
	VCSAG_LOGDBG_MSG(1, "Setting up resync location $dstloc, Filername: $filername, source location: $srcloc ");
        for (my $ii = $resync_retries; $ii > 0; $ii--){
                $output = $handle->invoke("snapmirror-resync", "destination-location",
                        $dstloc, "source-location", $srcloc);
                if ($output->results_errno != 0) {
                        $r = $output->results_reason();
                        $ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("I", "ONTAPI 'snapmirror-resync' failed for destination location $dstloc, source location $srcloc at filer $filername. Error $output->{'results_errno'}: $r. Retrying ...", 1097, "$dstloc", "$srcloc", "$filername", "$output->{'results_errno'}", "$r");
                        $ENV{"LANG"}="C";
                        $resync_failure = 1;
	VCSAG_LOGDBG_MSG(1, "sub snapmirror-resync  failed Error: $r  : $output->{'results_errno'}");
                        sleep $resync_interval;
                }else{
                        $resync_failure = 0;
                        last;
                }
        }

        if ($resync_failure == 1) {
                        $ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-resync' failed for destination location $dstloc, source location $srcloc at filer $filername. Error $output->{'results_errno'}: $r. Giving up ...", 1098, "$dstloc", "$srcloc", "$filername", "$output->{'results_errno'}", "$r");
                        $ENV{"LANG"}="C";
                        return 1;
        }


        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("N", "snapmirror resync initiated for destination location $dstloc", 1099, "$dstloc");
        $ENV{"LANG"}="C";

        return 0;
}


sub SnapMirrorSetSchedule {
        my ($self, $dstloc, $sched) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorSetSchedule");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
        VCSAG_LOG_MSG("I", "Setting up snapmirror schedule for location $dstloc, Filername: $filername, schedule: $sched ", 1100, "$dstloc","$filername" ,"$sched");

	VCSAG_LOGDBG_MSG(1, "Setting up snapmirror schedule for location $dstloc, Filername: $filername, schedule: $sched ");
        my $output = $handle->invoke("snapmirror-modify",
                "destination-location", $dstloc,
                "schedule", $sched);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-set-schedule' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1101, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
	VCSAG_LOGDBG_MSG(1, "sub snapmirror-modify failed Error: $r  : $output->{'results_errno'}");
                return 1;
        }

	VCSAG_LOGDBG_MSG(1, "sub snapmirror-modify success");
        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror schedule set for location $dstloc", 1102, "$dstloc");
        $ENV{"LANG"}="C";

        return 0;
}

sub SnapMirrorSetPolicy {
        my ($self, $dstloc, $policy) = @_;
        VCSAG_LOGDBG_MSG(1, "sub SnapMirrorSetPolicy");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
        VCSAG_LOG_MSG("I", "Setting up snapmirror policy for location $dstloc, Filername: $filername, policy: $policy ", 1103, "$dstloc","$filername" ,"$policy");

        VCSAG_LOGDBG_MSG(1, "Setting up snapmirror policy for location $dstloc, Filername: $filername, policy: $policy ");
        my $output = $handle->invoke("snapmirror-modify",
                "destination-location", $dstloc,
                "policy", $policy);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-set-policy' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1104, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
        VCSAG_LOGDBG_MSG(1, "sub snapmirror-modify failed Error: $r  : $output->{'results_errno'}");
                return 1;
        }

        VCSAG_LOGDBG_MSG(1, "sub snapmirror-modify success");
        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror policy set for location $dstloc", 1031, "$dstloc");
        $ENV{"LANG"}="C";

        return 0;
}

sub SnapshotDelete {
        my ($self, $dstloc) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapshotDelete");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};


	VCSAG_LOGDBG_MSG(1, "snapmirror-destroy for destination-location: $dstloc");
        my $output = $handle->invoke("snapmirror-destroy",
                "destination-location", $dstloc);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOGDBG_MSG(1, "snapmirror-destroy for destination-location: $dstloc, at filer $filername. Error $output->{'results_errno'}: $r");
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-destroy' failed for destination: $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1105, "$dstloc",  "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror-destroy for destination-location $dstloc at filer $filername", 1106, "$dstloc", "$filername");
        $ENV{"LANG"}="C";

        return 0;
}

#
# Release the previous snapmirror relationship on this filer.
# i,e inform snapmirror on this filer (i.e. original source location) that a
# direct mirror (i.e. original destination location) is no longer going to make
# snapmirror requests.
#

sub SnapMirrorRelease {
        my ($self, $dstloc) = @_;
	VCSAG_LOGDBG_MSG(1, "sub SnapMirrorRelease");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};

        my $output = $handle->invoke("snapmirror-release",
                "destination-location", $dstloc);

        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-release' failed for destination location $dstloc,  filer $filername. Error $output->{'results_errno'}: $r", 1107, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
		VCSAG_LOGDBG_MSG(1, "sub SnapMirrorRelease failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r");
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror release succeeded for destination location $dstloc at  $filername", 1108, "$dstloc", "$filername");
        $ENV{"LANG"}="C";

        return 0;
}


sub takeover {
        my ($self, $sm_info) = @_;
	VCSAG_LOGDBG_MSG(1, "sub takeover");

        my $flrname = $self->{FILER_NAME};
        my $dstloc = $sm_info->{DST_LOC};

        #
        # check if snapmirror state is 'snapmirrored'
       
        my @sminfolist = $self->SnapMirrorGetInfo();
        return 1 if (!@sminfolist);
        #
        # we expect to see only one entry in the snapmirror infolist.
        # this agent does not support multiple snapmirror destinations
        # for the same source volume
        #
       
        if (scalar(@sminfolist) != 1) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("I", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $flrname returned multiple status entries. Administrative intervention required", 1109, "$dstloc", "$flrname");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $new_sm_info = $sminfolist[0];
        return 0 if ($new_sm_info->{STATE} eq "broken-off");

        if ($new_sm_info->{STATE} ne "snapmirrored") {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Volume $dstloc is in state $new_sm_info->{STATE}", 1110, "$dstloc", "$new_sm_info->{STATE}");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $ret = $self->SnapMirrorQuiesce($dstloc);
        if ($ret) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to queisce destination location $dstloc on filer $flrname", 1111, "$dstloc", "$flrname");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ret = $self->SnapMirrorBreak($dstloc);
        if ($ret) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Failed to break destination location $dstloc on filer $flrname", 1112, "$dstloc", "$flrname");
                $ENV{"LANG"}="C";
                return 1;
        }

	return 0; 
        

}

sub GetVolumeType {

	 my ($self) = @_;
        VCSAG_LOGDBG_MSG(1, "sub GetVolumeType");

        my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
        my $volume = $self->{VOLUME};

	my $output = $self->VolumeStatusAPI($handle);
        if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'volume-get-iter' failed for volume $volume on filer $filername. Error $output->{'results_errno'}: $r", 1113, "$volume", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }
		
	 VCSAG_LOGDBG_MSG(1, "Get attributes list");
	 my $attrlist = $output->child_get("attributes-list");
	 if (! defined($attrlist)) {
		VCSAG_LOG_MSG("E","Failed to get attributes list", 1117);
		return 1;
	 }
	 VCSAG_LOGDBG_MSG(1, "Get volume attributes");
	 my $volattr = $attrlist->child_get("volume-attributes");
	 if (! defined($volattr)) {
		 VCSAG_LOG_MSG("E","Failed to get volume attributes", 1118);
		 return 1;
	 }
	 
	 VCSAG_LOGDBG_MSG(1, "Get volume id attributes");
	 my $volidattr = $volattr->child_get("volume-id-attributes");
	 if (! defined($volidattr)) {
		VCSAG_LOG_MSG("E","Failed to get volume id attributes", 1119);
		return 1;
	 }
	 
	 VCSAG_LOGDBG_MSG(1, "Get type");
	 my $type = $volidattr->child_get_string( "type" );
	 	 if (! defined($volidattr)) {
		VCSAG_LOG_MSG("E","Failed to get volume type.", 1120);
		return 1;
	 }
	 
	 VCSAG_LOGDBG_MSG(1, "type = $type");
	
	if(($type eq "rw") || ($type eq "RW")){
                VCSAG_LOG_MSG("I", "ONTAPI volume: $volume Type: $type on filer $filername ", 1114, "$volume", "$type","$filername");
		return 0;
	
	}  

	VCSAG_LOG_MSG("I", "ONTAPI volume: $volume Type: $type on filer $filername ", 1115, "$volume", "$type","$filername");
	return 1;
}


sub VolumeStatusAPI {
	my $self = shift;
	my $handle = shift;
	my $vserver = $self->{VSERVER_NAME};
	my $volume = $self->{VOLUME};
	VCSAG_LOGDBG_MSG(1, "sub GetVolumeStatus");
	
	my $xi = new NaElement("volume-get-iter");

	my $xque = new NaElement("query");
	$xi->child_add($xque);

	my $xattr = new NaElement("volume-attributes");
	$xque->child_add($xattr);

	my $xid = new NaElement("volume-id-attributes");
	$xattr->child_add($xid);
	
	VCSAG_LOGDBG_MSG(1, "GetVolumeStatus with volume name:  $volume, vserver name:  $vserver");
	$xid->child_add_string('name', $volume);
	$xid->child_add_string('owning-vserver-name', $vserver);

	return $handle->invoke_elem($xi);
}

sub ISReplicationHealthy {
        my ($self ) = @_;
        VCSAG_LOGDBG_MSG(1, "sub ISReplicationHealthy");
        my @sminfolist;
        my $filername = $self->{FILER_NAME};

        @sminfolist = $self->SnapMirrorGetInfo();
        return 1 if (!@sminfolist);

        if (scalar(@sminfolist) != 1) {
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for filer $filername returned multiple status entries. Administrative intervention required", 1117,  "$filername");
                $ENV{"LANG"}="C";
                return 1;
        }

        my $sm_info = $sminfolist[0];
        if ($sm_info->{HEALTHY} =~ /true/i ) {

                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("I", "Replication is-health state is true. Filer $filername", 1118, $filername);
                return 0;
        }
        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "Replication is-health state is not true. Filer $filername", 1119, $filername); 

        return 1;
}

 
1;




